Index: lisp/server.el
===================================================================
RCS file: /cvsroot/emacs/emacs/lisp/server.el,v
retrieving revision 1.113
diff -b -c -2 -r1.113 server.el
*** lisp/server.el	6 Jul 2006 22:48:16 -0000	1.113
--- lisp/server.el	26 Oct 2006 23:06:08 -0000
***************
*** 83,86 ****
--- 83,115 ----
    :group 'external)
  
+ (defcustom server-use-tcp nil
+   "If non-nil, use TCP sockets instead of local sockets."
+   :set #'(lambda (sym val)
+            (if (featurep 'make-network-process '(:family local))
+                (set-default sym val)
+              (unless load-in-progress
+                (message "Local sockets unsupported, using TCP sockets"))
+              (set-default sym t)))
+   :group 'server
+   :type 'boolean
+   :version "22.1")
+ 
+ (defcustom server-host nil
+   "The name or IP address to use as host address of the server process.
+ If set, the server accepts remote connections; otherwise it is local."
+   :group 'server
+   :type '(choice
+           (string :tag "Name or IP address")
+           (const :tag "Local" nil))
+   :version "22.1")
+ (put 'server-host 'risky-local-variable t)
+ 
+ (defcustom server-auth-dir "~/.emacs.d/server/"
+   "Directory for server authentication files."
+   :group 'server
+   :type 'directory
+   :version "22.1")
+ (put 'server-auth-dir 'risky-local-variable t)
+ 
  (defcustom server-visit-hook nil
    "*Hook run when visiting a file for the Emacs server."
***************
*** 101,104 ****
--- 130,136 ----
    "The current server process.")
  
+ (defvar server-auth-string nil
+   "The current server authentication string.")
+ 
  (defvar server-clients nil
    "List of current server clients.
***************
*** 232,242 ****
    (let ((attrs (file-attributes dir)))
      (unless attrs
!       (letf (((default-file-modes) ?\700)) (make-directory dir))
        (setq attrs (file-attributes dir)))
      ;; Check that it's safe for use.
      (unless (and (eq t (car attrs)) (eq (nth 2 attrs) (user-uid))
! 		 (zerop (logand ?\077 (file-modes dir))))
        (error "The directory %s is unsafe" dir))))
  
  ;;;###autoload
  (defun server-start (&optional leave-dead)
--- 264,285 ----
    (let ((attrs (file-attributes dir)))
      (unless attrs
!       (letf (((default-file-modes) ?\700)) (make-directory dir t))
        (setq attrs (file-attributes dir)))
      ;; Check that it's safe for use.
      (unless (and (eq t (car attrs)) (eq (nth 2 attrs) (user-uid))
!                  (or (eq system-type 'windows-nt)
!                      (zerop (logand ?\077 (file-modes dir)))))
        (error "The directory %s is unsafe" dir))))
  
+ (defun server-auth-string ()
+   (or server-auth-string
+       ;; If the authentication string does not exist, create it on the fly:
+       ;; it's a 64-byte string of random chars in the range `!'..`~'.
+       (setq server-auth-string
+ 	    (loop
+ 	       for i below 64
+ 	       collect (+ 33 (random 94)) into auth
+ 	       finally return (concat auth)))))
+ 
  ;;;###autoload
  (defun server-start (&optional leave-dead)
***************
*** 249,259 ****
  Prefix arg means just kill any existing server communications subprocess."
    (interactive "P")
    ;; kill it dead!
!   (if server-process
!       (condition-case () (delete-process server-process) (error nil)))
!   ;; Delete the socket files made by previous server invocations.
!   (condition-case ()
        (delete-file (expand-file-name server-name server-socket-dir))
!     (error nil))
    ;; If this Emacs already had a server, clear out associated status.
    (while server-clients
--- 292,304 ----
  Prefix arg means just kill any existing server communications subprocess."
    (interactive "P")
+   (when server-process
      ;; kill it dead!
!     (ignore-errors (delete-process server-process))
!     (ignore-errors
!       ;; Delete the socket or authentication files made by previous server invocations.
!       (if (eq (process-contact server-process :family) 'local)
            (delete-file (expand-file-name server-name server-socket-dir))
!         (setq server-auth-string nil)
!         (delete-file (expand-file-name server-name server-auth-dir)))))
    ;; If this Emacs already had a server, clear out associated status.
    (while server-clients
***************
*** 263,279 ****
    (unless leave-dead
      ;; Make sure there is a safe directory in which to place the socket.
!     (server-ensure-safe-dir server-socket-dir)
!     (if server-process
  	(server-log (message "Restarting server")))
      (letf (((default-file-modes) ?\700))
        (setq server-process
! 	    (make-network-process
! 	     :name "server" :family 'local :server t :noquery t
! 	     :service (expand-file-name server-name server-socket-dir)
! 	     :sentinel 'server-sentinel :filter 'server-process-filter
  	     ;; We must receive file names without being decoded.
  	     ;; Those are decoded by server-process-filter according
  	     ;; to file-name-coding-system.
! 	     :coding 'raw-text)))))
  
  ;;;###autoload
--- 308,342 ----
    (unless leave-dead
      ;; Make sure there is a safe directory in which to place the socket.
!     (server-ensure-safe-dir (if server-use-tcp server-auth-dir server-socket-dir))
!     (when server-process
        (server-log (message "Restarting server")))
      (letf (((default-file-modes) ?\700))
        (setq server-process
!             (apply #'make-network-process
!                    :name server-name
!                    :server t
!                    :noquery t
!                    :sentinel 'server-sentinel
!                    :filter 'server-process-filter
                     ;; We must receive file names without being decoded.
                     ;; Those are decoded by server-process-filter according
                     ;; to file-name-coding-system.
!                    :coding 'raw-text
!                    ;; The rest of the arguments depend on the kind of socket used
!                    (if server-use-tcp
!                        (list :family nil
!                              :service t
!                              :host (or server-host 'local)
!                              :plist '(:authenticated nil))
!                      (list :family 'local
!                            :service (expand-file-name server-name server-socket-dir)
!                            :plist '(:authenticated t))))))
!     (unless server-process (error "Could not start server process"))
!     (when server-use-tcp
!       (with-temp-file (expand-file-name server-name server-auth-dir)
!         (set-buffer-multibyte nil)
!         (setq buffer-file-coding-system 'no-conversion)
!         (insert (format-network-address (process-contact server-process :local))
!                 "\n" (server-auth-string))))))
  
  ;;;###autoload
***************
*** 290,301 ****
    (server-start (not server-mode)))
  
! (defun server-process-filter (proc string)
    "Process a request from the server to edit some files.
  PROC is the server process.  Format of STRING is \"PATH PATH PATH... \\n\"."
    (server-log string proc)
!   (let ((prev (process-get proc 'previous-string)))
      (when prev
        (setq string (concat prev string))
!       (process-put proc 'previous-string nil)))
    ;; If the input is multiple lines,
    ;; process each line individually.
--- 353,375 ----
    (server-start (not server-mode)))
  
! (defun* server-process-filter (proc string)
    "Process a request from the server to edit some files.
  PROC is the server process.  Format of STRING is \"PATH PATH PATH... \\n\"."
+   ;; First things first: let's check the authentication
+   (unless (process-get proc :authenticated)
+     (if (and (string-match "-auth \\(.*?\\)\n" string)
+              (string= (match-string 1 string) server-auth-string))
+         (progn
+           (setq string (substring string (match-end 0)))
+           (process-put proc :authenticated t)
+           (server-log "Authentication successful" proc))
+       (server-log "Authentication failed" proc)
+       (delete-process proc)
+       (return-from server-process-filter)))
    (server-log string proc)
!   (let ((prev (process-get proc :previous-string)))
      (when prev
        (setq string (concat prev string))
!       (process-put proc :previous-string nil)))
    ;; If the input is multiple lines,
    ;; process each line individually.
***************
*** 390,394 ****
    ;; Save for later any partial line that remains.
    (when (> (length string) 0)
!     (process-put proc 'previous-string string)))
  
  (defun server-goto-line-column (file-line-col)
--- 464,468 ----
    ;; Save for later any partial line that remains.
    (when (> (length string) 0)
!     (process-put proc :previous-string string)))
  
  (defun server-goto-line-column (file-line-col)
Index: lib-src/emacsclient.c
===================================================================
RCS file: /cvsroot/emacs/emacs/lib-src/emacsclient.c,v
retrieving revision 1.77
diff -b -c -2 -r1.77 emacsclient.c
*** lib-src/emacsclient.c	18 Jul 2006 16:33:45 -0000	1.77
--- lib-src/emacsclient.c	27 Oct 2006 00:06:01 -0000
***************
*** 27,35 ****
  #endif
  
  #undef signal
  
  #include <ctype.h>
  #include <stdio.h>
! #include <getopt.h>
  #ifdef HAVE_UNISTD_H
  #include <unistd.h>
--- 27,40 ----
  #endif
  
+ #ifdef WINDOWSNT
+ #define HAVE_SOCKETS
+ #define NO_SOCKETS_IN_FILE_SYSTEM
+ #endif
+ 
  #undef signal
  
  #include <ctype.h>
  #include <stdio.h>
! #include "getopt.h"
  #ifdef HAVE_UNISTD_H
  #include <unistd.h>
***************
*** 38,43 ****
  #ifdef VMS
  # include "vms-pwd.h"
! #else
  # include <pwd.h>
  #endif /* not VMS */
  
--- 43,52 ----
  #ifdef VMS
  # include "vms-pwd.h"
! #else /* not VMS */
! #ifdef WINDOWSNT
! # include <io.h>
! #else /* not WINDOWSNT */
  # include <pwd.h>
+ #endif /* not WINDOWSNT */
  #endif /* not VMS */
  
***************
*** 49,52 ****
--- 58,76 ----
  #endif
  
+ #define SEND_STRING(data) (send_to_emacs (s, (data)))
+ #define SEND_QUOTED(data) (quote_file_name (s, (data)))
+ 
+ #ifndef EXIT_SUCCESS
+ #define EXIT_SUCCESS 0
+ #endif
+ 
+ #ifndef EXIT_FAILURE
+ #define EXIT_FAILURE 1
+ #endif
+ 
+ #ifndef NO_RETURN
+ #define NO_RETURN
+ #endif
+ 
  /* Name used to invoke this program.  */
  char *progname;
***************
*** 68,71 ****
--- 92,98 ----
  char *socket_name = NULL;
  
+ /* If non-NULL, the filename of the authentication file.  */
+ char *server_file = NULL;
+ 
  void print_help_and_exit () NO_RETURN;
  
***************
*** 78,81 ****
--- 105,109 ----
    { "alternate-editor", required_argument, NULL, 'a' },
    { "socket-name",	required_argument, NULL, 's' },
+   { "server-file",	required_argument, NULL, 'f' },
    { "display",	required_argument, NULL, 'd' },
    { 0, 0, 0, 0 }
***************
*** 91,99 ****
  {
    alternate_editor = getenv ("ALTERNATE_EDITOR");
  
    while (1)
      {
        int opt = getopt_long (argc, argv,
! 			     "VHnea:s:d:", longopts, 0);
  
        if (opt == EOF)
--- 119,128 ----
  {
    alternate_editor = getenv ("ALTERNATE_EDITOR");
+   server_file = getenv ("EMACS_SERVER_FILE");
  
    while (1)
      {
        int opt = getopt_long (argc, argv,
! 			     "VHnea:s:f:d:", longopts, 0);
  
        if (opt == EOF)
***************
*** 115,118 ****
--- 144,151 ----
  	  break;
  
+ 	case 'f':
+ 	  server_file = optarg;
+ 	  break;
+ 
  	case 'd':
  	  display = optarg;
***************
*** 158,163 ****
  -e, --eval              Evaluate the FILE arguments as ELisp expressions\n\
  -d, --display=DISPLAY   Visit the file in the given display\n\
! -s, --socket-name=FILENAME\n\
                          Set the filename of the UNIX socket for communication\n\
  -a, --alternate-editor=EDITOR\n\
                          Editor to fallback to if the server is not running\n\
--- 191,202 ----
  -e, --eval		Evaluate the FILE arguments as ELisp expressions\n\
  -d, --display=DISPLAY	Visit the file in the given display\n\
! "
! #ifndef NO_SOCKETS_IN_FILE_SYSTEM
! "-s, --socket-name=FILENAME\n\
  			Set the filename of the UNIX socket for communication\n\
+ "
+ #endif
+ "-f, --server-file=FILENAME\n\
+ 			Set the filename of the TCP configuration file\n\
  -a, --alternate-editor=EDITOR\n\
  			Editor to fallback to if the server is not running\n\
***************
*** 167,178 ****
  }
  
  /* In NAME, insert a & before each &, each space, each newline, and
     any initial -.  Change spaces to underscores, too, so that the
     return value never contains a space.  */
- 
  void
! quote_file_name (name, stream)
       char *name;
-      FILE *stream;
  {
    char *copy = (char *) malloc (strlen (name) * 2 + 1);
--- 206,313 ----
  }
  
+ 
+ /*
+   Try to run a different command, or --if no alternate editor is
+   defined-- exit with an errorcode.
+ */
+ void
+ fail (argc, argv)
+      int argc;
+      char **argv;
+ {
+   if (alternate_editor)
+     {
+       int i = optind - 1;
+       execvp (alternate_editor, argv + i);
+       return;
+     }
+   else
+     {
+       exit (EXIT_FAILURE);
+     }
+ }
+ 
+ 
+ #if !defined (HAVE_SOCKETS)
+ 
+ int
+ main (argc, argv)
+      int argc;
+      char **argv;
+ {
+   fprintf (stderr, "%s: Sorry, the Emacs server is supported only\n",
+ 	   argv[0]);
+   fprintf (stderr, "on systems with Berkeley sockets.\n");
+ 
+   fail (argc, argv);
+ }
+ 
+ #else /* HAVE_SOCKETS */
+ 
+ #define AUTH_STRING_LENGTH   64
+ #define SEND_BUFFER_SIZE   4096
+ 
+ #ifdef WINDOWSNT
+ # include <winsock2.h>
+ #else
+ # include <sys/types.h>
+ # include <sys/socket.h>
+ # include <sys/un.h>
+ # include <sys/stat.h>
+ # include <errno.h>
+ #endif
+ 
+ extern char *strerror ();
+ extern int errno;
+ 
+ /* Buffer to accumulate data to send in TCP connections.  */
+ char send_buffer[SEND_BUFFER_SIZE + 1];
+ int sblen = 0;	/* Fill pointer for the send buffer.  */
+ 
+ /* Let's send the data to Emacs when either
+    - the data ends in "\n", or
+    - the buffer is full (but this shouldn't happen)
+    Otherwise, we just accumulate it.  */
+ void send_to_emacs (s, data)
+      SOCKET s;
+      char *data;
+ {
+   while (data)
+     {
+       int dlen = strlen (data);
+       if (dlen + sblen >= SEND_BUFFER_SIZE)
+ 	{
+ 	  int part = SEND_BUFFER_SIZE - sblen;
+ 	  strncpy (&send_buffer[sblen], data, part);
+ 	  data += part;
+ 	  sblen = SEND_BUFFER_SIZE;
+ 	}
+       else if (dlen)
+ 	{
+ 	  strcpy (&send_buffer[sblen], data);
+ 	  data = NULL;
+ 	  sblen += dlen;
+ 	}
+       else
+ 	break;
+ 
+       if (sblen == SEND_BUFFER_SIZE
+ 	  || (sblen > 0 && send_buffer[sblen-1] == '\n'))
+ 	{
+ 	  int sent = send (s, send_buffer, sblen, 0);
+ 	  if (sent != sblen)
+ 	    strcpy (send_buffer, &send_buffer[sent]);
+ 	  sblen -= sent;
+ 	}
+     }
+ }
+ 
  /* In NAME, insert a & before each &, each space, each newline, and
     any initial -.  Change spaces to underscores, too, so that the
     return value never contains a space.	 */
  void
! quote_file_name (s, name)
!      SOCKET s;
       char *name;
  {
    char *copy = (char *) malloc (strlen (name) * 2 + 1);
***************
*** 204,274 ****
    *q++ = 0;
  
!   fprintf (stream, "%s", copy);
  
    free (copy);
  }
  
! /* Like malloc but get fatal error if memory is exhausted.  */
  
! long *
! xmalloc (size)
!      unsigned int size;
  {
!   long *result = (long *) malloc (size);
!   if (result == NULL)
    {
!     perror ("malloc");
      exit (EXIT_FAILURE);
    }
!   return result;
  }
  
  /*
!   Try to run a different command, or --if no alternate editor is
!   defined-- exit with an errorcode.
  */
! void
! fail (argc, argv)
!      int argc;
!      char **argv;
  {
!   if (alternate_editor)
      {
!       int i = optind - 1;
!       execvp (alternate_editor, argv + i);
!       return;
      }
    else
      {
        exit (EXIT_FAILURE);
      }
- }
  
  
  
! #if !defined (HAVE_SOCKETS) || defined (NO_SOCKETS_IN_FILE_SYSTEM)
  
! int
! main (argc, argv)
!      int argc;
!      char **argv;
  {
!   fprintf (stderr, "%s: Sorry, the Emacs server is supported only\n",
! 	   argv[0]);
!   fprintf (stderr, "on systems with Berkeley sockets.\n");
  
!   fail (argc, argv);
! }
  
! #else /* HAVE_SOCKETS */
  
! #include <sys/types.h>
! #include <sys/socket.h>
! #include <sys/un.h>
! #include <sys/stat.h>
! #include <errno.h>
  
! extern char *strerror ();
! extern int errno;
  
  /* Three possibilities:
--- 339,463 ----
    *q++ = 0;
  
!   SEND_STRING (copy);
  
    free (copy);
  }
  
! #ifdef WINDOWSNT
! /* Wrapper to make WSACleanup a cdecl, as required by atexit().	 */
! void close_winsock ()
! {
!   WSACleanup ();
! }
! #endif /* WINDOWSNT */
  
! void initialize_sockets ()
  {
! #ifdef WINDOWSNT
!   WSADATA wsaData;
! 
!   /* Initialize the WinSock2 library.  */
!   if (WSAStartup (MAKEWORD (2, 0), &wsaData))
      {
!       fprintf (stderr, "%s: error initializing WinSock2", progname);
        exit (EXIT_FAILURE);
      }
! 
!   atexit (close_winsock);
! #endif /* WINDOWSNT */
  }
  
  /*
!  * Read the information needed to set up a TCP comm channel with
!  * the Emacs server: host, port and authentication string.
  */
! void get_server_config (server, authentication)
!      struct sockaddr_in *server;
!      char *authentication;
  {
!   FILE *config;
!   char dotted[32];
!   char *port;
! 
!   if (! (config = fopen (server_file, "rb")))
      {
!       fprintf (stderr, "%s: cannot read configuration file %s",
! 	       progname, server_file);
!       exit (EXIT_FAILURE);
!     }
! 
!   if (fgets (dotted, sizeof dotted, config)
!       && (port = strchr (dotted, ':')))
!     {
!       *port++ = '\0';
      }
    else
      {
+       fprintf (stderr, "%s: invalid configuration info", progname);
        exit (EXIT_FAILURE);
      }
  
+   server->sin_family = AF_INET;
+   server->sin_addr.s_addr = inet_addr (dotted);
+   server->sin_port = htons (atoi (port));
  
+   if (! fread (authentication, AUTH_STRING_LENGTH, 1, config))
+     {
+       fprintf (stderr, "%s: cannot read authentication info", progname);
+       exit (EXIT_FAILURE);
+     }
  
!   fclose (config);
! }
  
! SOCKET
! set_tcp_socket ()
  {
!   SOCKET s;
!   struct sockaddr_in server;
!   unsigned long c_arg = 0;
!   struct linger l_arg = {1, 1};
!   char auth_string[AUTH_STRING_LENGTH + 1];
  
!   initialize_sockets ();
  
!   get_server_config (&server, auth_string);
  
!   /*
!    * Open up an AF_INET socket
!    */
!   if ((s = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
!     {
!       fprintf (stderr, "%s: ", progname);
!       perror ("socket");
!       return INVALID_SOCKET;
!     }
  
!   /*
!    * Set up the socket
!    */
!   if (connect (s, (struct sockaddr *) &server, sizeof server) < 0)
!     {
!       fprintf (stderr, "%s: ", progname);
!       perror ("connect");
!       return INVALID_SOCKET;
!     }
! 
!   ioctlsocket (s, FIONBIO, &c_arg);
!   setsockopt (s, SOL_SOCKET, SO_LINGER, (char *) &l_arg, sizeof l_arg);
! 
!   /*
!    * Send the authentication
!    */
!   auth_string[AUTH_STRING_LENGTH] = '\0';
! 
!   SEND_STRING ("-auth ");
!   SEND_STRING (auth_string);
!   SEND_STRING ("\n");
! 
!   return s;
! }
! 
! #if !defined (NO_SOCKETS_IN_FILE_SYSTEM)
  
  /* Three possibilities:
***************
*** 292,317 ****
  }
  
! int
! main (argc, argv)
!      int argc;
!      char **argv;
  {
!   int s, i, needlf = 0;
!   FILE *out, *in;
    struct sockaddr_un server;
-   char *cwd, *str;
-   char string[BUFSIZ];
- 
-   progname = argv[0];
- 
-   /* Process options.  */
-   decode_options (argc, argv);
- 
-   if ((argc - optind < 1) && !eval)
-     {
-       fprintf (stderr, "%s: file name or argument required\n", progname);
-       fprintf (stderr, "Try `%s --help' for more information\n", progname);
-       exit (EXIT_FAILURE);
-     }
  
    /*
--- 481,489 ----
  }
  
! SOCKET
! set_local_socket ()
  {
!   SOCKET s;
    struct sockaddr_un server;
  
    /*
***************
*** 321,327 ****
    if ((s = socket (AF_UNIX, SOCK_STREAM, 0)) < 0)
      {
!       fprintf (stderr, "%s: ", argv[0]);
        perror ("socket");
!       fail (argc, argv);
      }
  
--- 493,499 ----
    if ((s = socket (AF_UNIX, SOCK_STREAM, 0)) < 0)
      {
!       fprintf (stderr, "%s: ", progname);
        perror ("socket");
!       return INVALID_SOCKET;
      }
  
***************
*** 353,357 ****
        {
  	fprintf (stderr, "%s: socket-name %s too long",
! 		 argv[0], socket_name);
  	exit (EXIT_FAILURE);
        }
--- 525,529 ----
        {
  	fprintf (stderr, "%s: socket-name %s too long",
! 		 progname, socket_name);
  	exit (EXIT_FAILURE);
        }
***************
*** 388,392 ****
  		  {
  		    fprintf (stderr, "%s: socket-name %s too long",
! 			     argv[0], socket_name);
  		    exit (EXIT_FAILURE);
  		  }
--- 560,564 ----
  		  {
  		    fprintf (stderr, "%s: socket-name %s too long",
! 			     progname, socket_name);
  		    exit (EXIT_FAILURE);
  		  }
***************
*** 408,412 ****
  	   {
  	     fprintf (stderr, "%s: Invalid socket owner\n", argv[0]);
! 	     fail (argc, argv);
  	   }
  	 break;
--- 580,584 ----
  	  {
  	    fprintf (stderr, "%s: Invalid socket owner\n", argv[0]);
! 	    return INVALID_SOCKET;
  	  }
  	break;
***************
*** 418,427 ****
  		    "%s: can't find socket; have you started the server?\n\
  To start the server in Emacs, type \"M-x server-start\".\n",
! 		    argv[0]);
  	 else
  	   fprintf (stderr, "%s: can't stat %s: %s\n",
! 		    argv[0], server.sun_path, strerror (saved_errno));
! 	 fail (argc, argv);
! 	 break;
         }
    }
--- 590,598 ----
  		   "%s: can't find socket; have you started the server?\n\
  To start the server in Emacs, type \"M-x server-start\".\n",
! 		   progname);
  	else
  	  fprintf (stderr, "%s: can't stat %s: %s\n",
! 		   progname, server.sun_path, strerror (saved_errno));
! 	return INVALID_SOCKET;
        }
    }
***************
*** 430,458 ****
        < 0)
      {
!       fprintf (stderr, "%s: ", argv[0]);
        perror ("connect");
!       fail (argc, argv);
      }
  
!   /* We use the stream OUT to send our command to the server.  */
!   if ((out = fdopen (s, "r+")) == NULL)
      {
!       fprintf (stderr, "%s: ", argv[0]);
!       perror ("fdopen");
!       fail (argc, argv);
      }
  
!   /* We use the stream IN to read the response.
!      We used to use just one stream for both output and input
!      on the socket, but reversing direction works nonportably:
!      on some systems, the output appears as the first input;
!      on other systems it does not.  */
!   if ((in = fdopen (s, "r+")) == NULL)
      {
!       fprintf (stderr, "%s: ", argv[0]);
!       perror ("fdopen");
!       fail (argc, argv);
      }
  
  #ifdef HAVE_GETCWD
    cwd = getcwd (string, sizeof string);
--- 601,654 ----
        < 0)
      {
!       fprintf (stderr, "%s: ", progname);
        perror ("connect");
!       return INVALID_SOCKET;
      }
  
!   return s;
! }
! #endif /* ! NO_SOCKETS_IN_FILE_SYSTEM */
! 
! SOCKET
! set_socket ()
! {
!   if (server_file)
!     return set_tcp_socket ();
!   else
! #ifndef NO_SOCKETS_IN_FILE_SYSTEM
!     return set_local_socket ();
! #else
      {
!       fprintf (stderr, "%s: missing --server-file option", progname);
!       exit (EXIT_FAILURE);
      }
+ #endif
+ }
+ 
+ int
+ main (argc, argv)
+      int argc;
+      char **argv;
+ {
+   SOCKET s;
+   int i, rl, needlf = 0;
+   char *cwd;
+   char string[BUFSIZ+1];
+ 
+   progname = argv[0];
  
!   /* Process options.  */
!   decode_options (argc, argv);
! 
!   if ((argc - optind < 1) && !eval)
      {
!       fprintf (stderr, "%s: file name or argument required\n", progname);
!       fprintf (stderr, "Try `%s --help' for more information\n", progname);
!       exit (EXIT_FAILURE);
      }
  
+   if ((s = set_socket ()) == INVALID_SOCKET)
+     fail (argc, argv);
+ 
  #ifdef HAVE_GETCWD
    cwd = getcwd (string, sizeof string);
***************
*** 465,472 ****
  
  #ifdef HAVE_GETCWD
!       fprintf (stderr, "%s: %s (%s)\n", argv[0],
  	       "Cannot get current working directory", strerror (errno));
  #else
!       fprintf (stderr, "%s: %s (%s)\n", argv[0], string, strerror (errno));
  #endif
        fail (argc, argv);
--- 661,668 ----
  
  #ifdef HAVE_GETCWD
!       fprintf (stderr, "%s: %s (%s)\n", progname,
  	       "Cannot get current working directory", strerror (errno));
  #else
!       fprintf (stderr, "%s: %s (%s)\n", progname, string, strerror (errno));
  #endif
        fail (argc, argv);
***************
*** 474,487 ****
  
    if (nowait)
!     fprintf (out, "-nowait ");
  
    if (eval)
!     fprintf (out, "-eval ");
  
    if (display)
      {
!       fprintf (out, "-display ");
!       quote_file_name (display, out);
!       fprintf (out, " ");
      }
  
--- 670,683 ----
  
    if (nowait)
!     SEND_STRING ("-nowait ");
  
    if (eval)
!     SEND_STRING ("-eval ");
  
    if (display)
      {
!       SEND_STRING ("-display ");
!       SEND_QUOTED (display);
!       SEND_STRING (" ");
      }
  
***************
*** 498,531 ****
  	      if (*p != 0)
  		{
! 		  quote_file_name (cwd, out);
! 		  fprintf (out, "/");
  		}
  	    }
  	  else if (*argv[i] != '/')
  	    {
! 	      quote_file_name (cwd, out);
! 	      fprintf (out, "/");
  	    }
  
! 	  quote_file_name (argv[i], out);
! 	  fprintf (out, " ");
  	}
      }
    else
      {
!       while ((str = fgets (string, BUFSIZ, stdin)))
  	{
! 	  quote_file_name (str, out);
  	}
!       fprintf (out, " ");
      }
  
!   fprintf (out, "\n");
!   fflush (out);
  
    /* Maybe wait for an answer.   */
!   if (nowait)
!     return EXIT_SUCCESS;
! 
    if (!eval)
      {
--- 694,734 ----
  	      if (*p != 0)
  		{
! 		  SEND_QUOTED (cwd);
! 		  SEND_STRING ("/");
  		}
  	    }
+ #ifndef WINDOWSNT
  	  else if (*argv[i] != '/')
+ #else
+ 	  else if ((*argv[i] != '/')
+ 		   /* Absolute paths can also start with backslash
+ 		      or drive letters.	 */
+ 		   && (*argv[i] != '\\')
+ 		   && (!islower (tolower (*argv[i]))
+ 		       || (argv[i][1] != ':')))
+ #endif
  	    {
! 	      SEND_QUOTED (cwd);
! 	      SEND_STRING ("/");
  	    }
  
! 	  SEND_QUOTED (argv[i]);
! 	  SEND_STRING (" ");
  	}
      }
    else
      {
!       while (fgets (string, BUFSIZ, stdin))
  	{
! 	  SEND_QUOTED (string);
  	}
!       SEND_STRING (" ");
      }
  
!   SEND_STRING ("\n");
  
    /* Maybe wait for an answer.	 */
!   if (!nowait)
!     {
        if (!eval)
  	{
***************
*** 536,545 ****
  
    /* Now, wait for an answer and print any messages.  */
!   while ((str = fgets (string, BUFSIZ, in)))
      {
        if (needlf == 2)
  	printf ("\n");
!       printf ("%s", str);
!       needlf = str[0] == '\0' ? needlf : str[strlen (str) - 1] != '\n';
      }
  
--- 739,749 ----
  
        /* Now, wait for an answer and print any messages.  */
!       while ((rl = recv (s, string, BUFSIZ, 0)) > 0)
  	{
+ 	  string[rl] = '\0';
  	  if (needlf == 2)
  	    printf ("\n");
! 	  printf ("%s", string);
! 	  needlf = string[0] == '\0' ? needlf : string[strlen (string) - 1] != '\n';
  	}
  
***************
*** 547,551 ****
--- 751,757 ----
  	printf ("\n");
        fflush (stdout);
+     }
  
+   closesocket (s);
    return EXIT_SUCCESS;
  }
Index: lib-src/makefile.w32-in
===================================================================
RCS file: /cvsroot/emacs/emacs/lib-src/makefile.w32-in,v
retrieving revision 2.46
diff -b -c -2 -r2.46 makefile.w32-in
*** lib-src/makefile.w32-in	9 Oct 2006 19:57:43 -0000	2.46
--- lib-src/makefile.w32-in	26 Oct 2006 22:23:12 -0000
***************
*** 21,25 ****
  #
  
! ALL = make-docfile hexl ctags etags movemail ebrowse sorted-doc digest-doc
  
  .PHONY: $(ALL)
--- 21,25 ----
  #
  
! ALL = make-docfile hexl ctags etags movemail ebrowse sorted-doc digest-doc emacsclient
  
  .PHONY: $(ALL)
***************
*** 33,37 ****
  #		$(BLD)/server.exe	\
  #		$(BLD)/emacstool.exe	\
- #		$(BLD)/emacsclient.exe	\
  #		$(BLD)/cvtmail.exe	\
  
--- 33,36 ----
***************
*** 60,63 ****
--- 59,63 ----
  sorted-doc:	$(BLD) $(BLD)/sorted-doc.exe
  digest-doc:	$(BLD) $(BLD)/digest-doc.exe
+ emacsclient:	$(BLD) $(BLD)/emacsclient.exe
  
  test-distrib:	$(BLD) $(BLD)/test-distrib.exe
***************
*** 75,78 ****
--- 75,91 ----
  		$(LINK) $(LINK_OUT)$@ $(LINK_FLAGS) $(MOVEMAILOBJS) $(WSOCK32) $(LIBS)
  
+ ECLIENT_CFLAGS = -DWINDOWSNT -DHAVE_GETCWD -DHAVE_STRERROR -c
+ ECLIENTOBJS =	$(BLD)/emacsclient.$(O) \
+ 		$(BLD)/getopt.$(O) \
+ 		$(BLD)/getopt1.$(O) \
+ 		$(BLD)/ntlib.$(O)
+ 
+ $(BLD)/emacsclient.exe:		$(ECLIENTOBJS)
+ # put wsock32.lib before $(LIBS) to ensure we don't link to ws2_32.lib
+ 		$(LINK) $(LINK_OUT)$@ $(LINK_FLAGS) $(ECLIENTOBJS) $(WSOCK32) $(LIBS)
+ 
+ $(BLD)/emacsclient.$(O):	emacsclient.c
+ 		$(CC) $(ECLIENT_CFLAGS) $(CC_OUT)$@ emacsclient.c
+ 
  ETAGSOBJ      = $(BLD)/etags.$(O) \
  		$(BLD)/getopt.$(O) \
***************
*** 297,300 ****
--- 310,314 ----
  		$(CP) $(BLD)/sorted-doc.exe $(INSTALL_DIR)/bin
  		$(CP) $(BLD)/digest-doc.exe $(INSTALL_DIR)/bin
+ 		$(CP) $(BLD)/emacsclient.exe $(INSTALL_DIR)/bin
  		- mkdir "$(INSTALL_DIR)/etc"
  		$(CP) $(DOC) $(INSTALL_DIR)/etc
